#include <signal.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#include <net/ethernet.h>
#include <net/if_arp.h>
#include <net/route.h>
#include <fcntl.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <string.h>
#include <sys/times.h>
#include <sys/wait.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/timeb.h>
#include <errno.h>
#include <sys/file.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <ctype.h>

//#define  __DEBUG__

unsigned long   GetTicks()
{
        double t;
        timeb tb;
        ftime(&tb);
        t = tb.time;
        t *= 1000.0;
        t += tb.millitm;
        return (unsigned long) t;
}


#define __ARP_REQUEST           1
#define __ARP_REPLY             2
#define __ETHERTYPE_IP          0x0800
#define __ETHERTYPE_ARP         0x0806
#define __ARPTYPE_ETHER         1

class ARPReq {
public:
   unsigned char     DestHW[6];
   unsigned char     SrcHW[6];
   unsigned short    EtherType;
   unsigned short    ARPType;          //Hardware type
   unsigned short    ProtoType;        //Protocol type
   unsigned char     hLength;          //Hardware address length
   unsigned char     pLength;          //Protocol address length
   unsigned short    op;               //Operator
   unsigned char     SrcAddress[6];    //Sendors Hardware address
   unsigned char     SrcIP[4];         //Sendors IP address
   unsigned char     DestAddress[6];   //Target Hardware address
   unsigned char     DestIP[4];        //Target IP address
   unsigned char     mtu[1500];

   int   Recv(int,sockaddr &);
   void  Send(int,sockaddr &);
};

class ProxyRule {
public:
   static ProxyRule *      first;
   ProxyRule *             next;

   unsigned long  invert;
   unsigned long  network;
   unsigned long  netmask;

   ProxyRule(unsigned long,unsigned long,unsigned long);

   int            Test(unsigned long,unsigned long);
   static int     Pass(unsigned long SrcIP,unsigned long DestIP);
};

class ProxyRecord;
class ARPInterface {
public:
   static ARPInterface *   first;
   ARPInterface *          next;

   unsigned short    res;                 //doesn't matter (0)
   char              Interface[16];       //interface name
   unsigned char     HWAddress[6];

   ARPInterface(char *);
   void           Send(int,ARPReq &);
   static int     Relay(int,sockaddr &,ARPReq &);
};

class ProxyRecord {
public:
   static ProxyRecord * first;
   static ProxyRecord * last;
   ProxyRecord *        prev;
   ProxyRecord *        next;

   unsigned long     StartTime;
   unsigned long     SrcIP;
   unsigned char     SrcAddress[6];
   ARPInterface &    Interface;

   ProxyRecord(ARPInterface &,ARPReq &);
   ~ProxyRecord();
   static int     Respond(int,ARPReq &);

   static int     Select(int,unsigned long);
   static void    Main();
};

/***/

ProxyRule *       ProxyRule::first = 0;
ARPInterface *    ARPInterface::first = 0;
ProxyRecord *     ProxyRecord::first = 0;
ProxyRecord *     ProxyRecord::last = 0;

/********************************** ARPReq:: ******************************/

int   ARPReq::Recv(int s,sockaddr &sa)
{
   socklen_t n;
   int c;

   n = sizeof(sa);
   c = recvfrom(s,this,1500,0,&sa,&n);

   if(htons(EtherType) != __ETHERTYPE_ARP) return 0;
   if(htons(ARPType) != __ARPTYPE_ETHER) return 0;
   if(htons(ProtoType) != __ETHERTYPE_IP) return 0;
   if(hLength != 6) return 0;
   if(pLength != 4) return 0;

   return 1;
}

void  ARPReq::Send(int s,sockaddr &sa)
{
   memcpy(DestHW,DestAddress,6);
   memcpy(SrcHW,SrcAddress,6);
   if(sendto(s,this,60,0,&sa,sizeof(sa)) <= 0) {
      printf(" (SNDERR)");
   };
}

/****************************** ProxyRule::* ******************************/

ProxyRule::ProxyRule(unsigned long a,unsigned long b,unsigned long c)
{
   next = first;
   first = this;

   invert = a;
   network = b;
   netmask = c;
}

int   ProxyRule::Test(unsigned long SrcIP,unsigned long DestIP)
{
   if((SrcIP & netmask) == (network & netmask)) {
      if(SrcIP == network) return 0;
      if(invert) return 0;
   }
   else {
      if(!invert) return 0;
   };

   if((DestIP & netmask) == (network & netmask)) {
      if(DestIP == network) return 0;
      if(invert) return 0;
   }
   else {
      if(!invert) return 0;
   };

   return 1;
}

int   ProxyRule::Pass(unsigned long SrcIP,unsigned long DestIP)
{
   ProxyRule *o;

   o = first;
   while(o) {
      if(!o[0].Test(SrcIP,DestIP)) return 0;
      o = o[0].next;
   };

   return 1;
}

/****************************** ARPInterface:: ****************************/

ARPInterface::ARPInterface(char *iface)
{
   next = first;
   first = this;

   strncpy(Interface,iface,16);
   struct ifreq ifr;
   int s = socket(PF_INET,SOCK_DGRAM,IPPROTO_UDP);
   strncpy(ifr.ifr_name,iface,sizeof(ifr.ifr_name));
   if(ioctl(s,SIOCGIFHWADDR,&ifr) == 0) {
      memcpy(HWAddress,ifr.ifr_hwaddr.sa_data,6);
   };
   close(s);
}

void  ARPInterface::Send(int s,ARPReq &req)
{
   memcpy(req.SrcAddress,HWAddress,6);
   res = 0;
   req.Send(s,((sockaddr *) &res) [0]);
}

int   ARPInterface::Relay(int s,sockaddr &sa,ARPReq &req)
{
   if(htons(req.op) != __ARP_REQUEST) return 0;

   memset(req.DestAddress,255,6);

   ARPInterface *o = first;
   while(o) {
      //Sanity checks
      //1..The source hardware address must not be local
      //2..The receiving interface must be registered with this service
      if(memcmp(o[0].HWAddress,req.SrcHW,6) == 0) return 0;
      if(strcmp(o[0].Interface,sa.sa_data) == 0) break;
      o = o[0].next;
   };
   if(o == 0) return 0;

   if(!ProxyRule::Pass(((unsigned long *) req.SrcIP) [0],
                       ((unsigned long *) req.DestIP) [0])) return 0;

   new ProxyRecord(o[0],req);

   o = first;
   while(o) {
      if(strcmp(o[0].Interface,sa.sa_data) != 0) {
         o[0].Send(s,req);
      };
      o = o[0].next;
   };

   return 1;
}

/****************************** ProxyRecord:: *****************************/

ProxyRecord::ProxyRecord(ARPInterface &iface,ARPReq &req) : Interface(iface)
{
   prev = last;
   next = 0;
   if(last) last[0].next = this;
   else first = this;
   last = this;

   printf("Relay: ");
   printf("%s, ",inet_ntoa(((in_addr *) req.SrcIP)[0]) );
   printf("%s ",inet_ntoa(((in_addr *) req.DestIP)[0]) );
   puts("");

   StartTime = GetTicks();
   SrcIP = ((unsigned long *) req.SrcIP) [0];
   memcpy(SrcAddress,req.SrcAddress,6);
}

ProxyRecord::~ProxyRecord()
{
   if(prev) prev[0].next = next;
   else first = next;
   if(next) next[0].prev = prev;
   else last = prev;
}

int   ProxyRecord::Respond(int s,ARPReq &req)
{
   if(htons(req.op) != __ARP_REPLY) return 0;

   ProxyRecord *o = first;
   while(o) {
      if(((unsigned long *) req.DestIP) [0] == o[0].SrcIP) break;
      o = o[0].next;
   };

   if(o == 0) return 0;

   printf("Reverse relay: ");
   printf("%s, ",inet_ntoa(((in_addr *) req.SrcIP)[0]) );
   printf("%s ",inet_ntoa(((in_addr *) req.DestIP)[0]) );
   puts("");

   memcpy(req.DestAddress,o[0].SrcAddress,6);
   o[0].Interface.Send(s,req);
   delete o;
   return 1;
}

int   ProxyRecord::Select(int s,unsigned long w)
{
   fd_set fd;
   int c;

   FD_ZERO(&fd);
   FD_SET(s,&fd);

   if(w == 0) c = select(s+1,&fd,0,0,0);
   else {
      timeval tv;
      tv.tv_sec = w / 1000UL;
      tv.tv_usec = (w % 1000UL) * 1000UL;
      c = select(s+1,&fd,0,0,&tv);
   };

   if(c < 0) exit(0);
   return c;
}

void  ProxyRecord::Main()
{
   unsigned long Timer;
   ARPReq req;
   sockaddr sa;
   int s;

   s = -1;
   if(s == -1) s = socket(AF_PACKET,SOCK_PACKET,htons(ETH_P_ALL));
   if(s == -1) s = socket(AF_INET,SOCK_PACKET,htons(ETH_P_ALL));
   if(s == -1) exit(0);

   while(1) {
      Timer = 5000;
      while(first != 0 && (Timer = GetTicks() - first[0].StartTime) >= 5000) {
         delete first;
         Timer = 5000;
      };

      if(!Select(s,5000 - Timer)) continue;
      if(!req.Recv(s,sa)) continue;

      if(ARPInterface::Relay(s,sa,req)) continue;
      if(Respond(s,req)) continue;
   };
}

main(int argc,char *argv[])
{
   char netaddr[64];

   if(argc < 2) {
      puts("Syntax: proxyarp [[!]net|host[/netmask]] ... iface1 iface2 ... ifacen");
      return 2;
   };

   int c = 1;
   while(c < argc) {
      if(strlen(argv[c]) < 64) {
         strcpy(netaddr,argv[c]);
         char *mask;
         char *net;

         mask = strchr(netaddr,'/');
         if(mask) {
            mask[0] = 0;
            mask++;
         }
         else mask = "255.255.255.255";

         net = netaddr;
         if(*net == '!') net++;
         if(inet_addr(net) != 0xFFFFFFFF) {
            if(*netaddr == '!') {
               new ProxyRule(1,inet_addr(net),inet_addr(mask));
            }
            else {
               new ProxyRule(0,inet_addr(net),inet_addr(mask));
            };

            c++;
            continue;
         };
      };

      new ARPInterface(argv[c]);
      c++;
   };

#ifndef __DEBUG__
   if(fork() <= 0) {
      close(0);
      close(1);
      close(2);
      setsid();
#endif
      ProxyRecord::Main();
#ifndef __DEBUG__
   };
#endif
   return 0;
}
